// SPDX-FileCopyrightText: 2024 Manuel Alcaraz Zambrano <manuelalcarazzam@gmail.com>
// SPDX-License-Identifier: LGPL-2.1-only OR LGPL-3.0-only OR LicenseRef-KDE-Accepted-LGPL

#pragma once

#include <QPluginLoader>

#include <KAboutComponent>
#include <KAboutData>
#include <KAboutPerson>
#include <KAutoSaveFile>
#include <KBackup>
#include <KCompositeJob>
#include <KCoreAddons>
#include <KDirWatch>
#include <KFileSystemType>
#include <KFileUtils>
#include <KFormat>
#include <KFuzzyMatcher>
#include <KJob>
#include <KJobTrackerInterface>
#include <KJobUiDelegate>
#include <KJsonUtils>
#include <KLibexec>
#include <KListOpenFilesJob>
#include <KMacroExpander>
#include <KMemoryInfo>
#include <KNetworkMounts>
#include <KOSRelease>
#include <KPluginFactory>
#include <KPluginMetaData>
#include <KProcess>
#include <KProcessList>
#include <KRandom>
#include <KRuntimePlatform>
#include <KSandbox>
#include <KSharedDataCache>
#include <KShell>
#include <KSignalHandler>
#include <KStringHandler>
#include <KTextToHTML>
#include <KUrlMimeData>
#include <KUser>
